-- Copyright   :  (c) Vidas Daudaravicius 2012
-- License     :  BSD3-style (see LICENSE)
-- Version     : 0.1


module Ngram where

import qualified System.IO as IO
import qualified Data.Map as M
import Files
import DeepSeq
import Data.Int
import Data.Char
import qualified Data.ByteString.Char8 as BS
import qualified Data.Binary as Bin

instance NFData (M.Map a b)
instance NFData (BS.ByteString)

------------------------------------------------------
tfidf d f n = f * log ( (n-d+1) / (d+1) )
tfidfNorm d docLen avgDocLen f n = (f*avgDocLen/docLen) * log ( (n-d+1) / (d+1) )


------------------------------------------------------
--------------        Dict reading         -----------
------------------------------------------------------
------------------------------------------------------



readDictBS'' file = do
                    text <- BS.readFile file
                    ls <- return $!! tail $ BS.lines text
                    return $!! map (\ln -> force (parseLineUTF8 ln)) ls
                    where
                       parseLineUTF8 ln = (w,read (BS.unpack f)::Double)
                                           where 
                                              (w,f) = BS.break (=='\t') ln


fromDictToBinDict file = do
                          dict <- readDict'' file
                          Bin.encodeFile (file++".bin") dict

writeBinDict file text = Bin.encodeFile file text

readBinDict file = Bin.decodeFile file    --- :: [BS.ByteString,Double]

readDict'' file = do
                    hIn <- IO.openFile file IO.ReadMode
                    text <- BS.hGetContents hIn
                    ls <- return $!! tail $ BS.lines text
                    IO.hClose hIn
                    return $!! map (\ln -> force (parseLineUTF8 ln)) ls
                    where
                       parseLineUTF8 ln = (w,read (BS.unpack f)::Double)
                                           where 
                                              (w,f) = BS.break (=='\t') ln

readDict'''' file = do
                    hIn <- IO.openFile file IO.ReadMode
                    ln <- BS.hGetLine hIn
                    d <- loadDict hIn M.empty
                    IO.hClose hIn
                    return d


--readDictUTF8 :: String -> IO [(String,Double)]
readDictUTF8 file = do
                    hIn <- IO.openFile file IO.ReadMode
                    text <- BS.hGetContents hIn
                    ls <- return $!! tail $ BS.lines text
                    IO.hClose hIn
                    return $!! map (\ln -> force $ parseLineUTF8 ln) $! ls
                    where
                       parseLineUTF8 ln = (w,read (BS.unpack f)::Double)
                                           where 
                                              (w,f) = BS.break (=='\t') ln

------------------------------------------------------
readDict' file ws = do
                    text <- BS.readFile file
                    ls <- return (BS.lines text)
                    docs <- return (read (BS.unpack $ head ls) :: Double)
                    return $!! (M.fromAscList  $ loadDict' ws (tail ls))
------------------------------------------------------
loadDict' [] ls = []
loadDict' ws [] = ws
loadDict' wss@((w',f'):ws) xss@(x:xs)
--          | w'==w  = (w,(f+f')): (force $ loadDict' ws xs)
          | w'==w  = (w,f): (loadDict' ws xs)
          | w'<w   = (w',f'): (loadDict' ws xss)
          | otherwise = (w,f):loadDict' wss xs
                  where
                      f = read (BS.unpack fr)::Double
                      (w,fr) = BS.break (=='\t') x

------------------------------------------------------
readDictDocs file = do
                 text <- IO.readFile file
                 ls <- return (lines text)
                 docs <- return (read (head ls) :: Double)
                 return docs
------------------------------------------------------
------------------------------------------------------
--------------Unigram frequency dictionary -----------
------------------------------------------------------
------------------------------------------------------
--readDocument :: Num a => FilePath -> IO (M.Map [Char] a)
readDocument fileName = do  text <- BS.readFile(fileName)
                            return $!! (makeDictFromText (BS.lines text))
------------------------------------------------------
--makeDictFromText :: Num a => [[Char]] -> M.Map [Char] a
makeDictFromText [] = M.empty
makeDictFromText (line:lines) =
            if BS.length line == 0
               then makeDictFromText lines
               else makeDictFromLine ( [(BS.pack "%.")] ++ (BS.words line) ++ [(BS.pack ".%")] ) $! makeDictFromText lines
------------------------------------------------------
makeDictFromLine []      uni  = uni
makeDictFromLine (x:rem) uni  = M.insertWith' (+) x 1 $!! makeDictFromLine rem uni
------------------------------------------------------
writeDict file uni docs = do   hFile <- IO.openFile file IO.WriteMode
                               BS.hPutStrLn hFile $ BS.pack $ show docs
                               printDict hFile docs $! M.toList $!! uni
                               IO.hClose hFile
------------------------------------------------------
printDict h n [] = return ()
printDict h n ((w,f):ls) = do  BS.hPutStrLn h (BS.concat [w,BS.pack "\t", BS.pack (show f)])
                               printDict h n ls
------------------------------------------------------
parseLine ln = (w,f')
                where
                  (w,f) = BS.break (=='\t') ln
                  f' = read (BS.unpack f)::Double
------------------------------------------------------
loadDict hIn d = do
                   end <- IO.hIsEOF hIn
                   if end
                      then return d
                      else do
                             line <- BS.hGetLine hIn
                             let (w,f) = parseLine line
                             loadDict hIn $!! M.insert w f d

------------------------------------------------------
------------------------------------------------------
--------------------Bigram dictionary ----------------
------------------------------------------------------
------------------------------------------------------
writeDictBi file bi docs = do
                     hFile <- IO.openFile file IO.WriteMode
                     BS.hPutStrLn hFile $ BS.pack $ show docs
                     printDictBi hFile $! M.toList $! bi
                     IO.hClose hFile
------------------------------------------------------
makeDictBi dir ext file = do
               files <- getFiles [dir] ext
               bi    <- readDocumentsBi files M.empty
               writeDictBi file bi (fromIntegral (length files) :: Double)
------------------------------------------------------
readDocumentBi fileName = do
                         text <- BS.readFile fileName
                         return $!! (makeDictFromTextBi (BS.lines text))
------------------------------------------------------
--makeDictFromTextBi :: [BS.ByteString] -> M.Map BS.ByteString (M.Map BS.ByteString Int)
makeDictFromTextBi [] = M.empty
makeDictFromTextBi (line:lines) =
               if BS.length line == 0 
                   then  makeDictFromTextBi lines
                   else  makeDictFromLineBi ( [BS.pack "%."] ++ (BS.words line) ++ [BS.pack ".%"] ) $! makeDictFromTextBi lines
------------------------------------------------------
makeDictFromLineBi::[BS.ByteString] -> M.Map BS.ByteString (M.Map BS.ByteString Int) -> M.Map BS.ByteString (M.Map BS.ByteString Int)
makeDictFromLineBi (x:y:rem) bi | BS.length x == 0    = makeDictFromLineBi (y:rem) bi
                                | BS.length y == 0    = makeDictFromLineBi (x:rem) bi
                                | otherwise           = M.insertWith' (\m n -> M.insertWith (+) y 1 n) x (M.singleton y 1) $! makeDictFromLineBi (y:rem) bi
makeDictFromLineBi _        bi  = bi
------------------------------------------------------
readDocumentsBi []           dict = return (dict)
readDocumentsBi (file:files) dict = do
                                      bi2 <- readDocumentBi file 
                                      putStrLn ((show (length files)) ++ "\t" ++ (show (M.size bi2)) ++ "\t" ++ file)
                                      readDocumentsBi files $!! (M.unionWith (\m n -> M.unionWith (+) m n) dict  bi2)
------------------------------------------------------
printDictBi' h w1 [] = return ()
printDictBi' h w1 ((w2,f):ls) = do
                          BS.hPutStrLn h $ BS.concat [w1,BS.pack " ", w2,BS.pack "\t",BS.pack $ show f]
                          printDictBi' h w1 ls
------------------------------------------------------
printDictBi h [] = return ()
printDictBi h ((w,f):ls) = do
                          printDictBi' h w (M.toList f)
                          printDictBi h ls

